/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.jameica.qrscanner.gui.dialog.threads;

import java.awt.Dimension;
import java.net.URL;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.TimeoutException;

import org.apache.commons.lang.NullArgumentException;
import org.apache.commons.lang.StringUtils;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;

import com.github.sarxos.webcam.Webcam;
import com.github.sarxos.webcam.WebcamDevice;
import com.github.sarxos.webcam.WebcamException;
import com.github.sarxos.webcam.ds.buildin.WebcamDefaultDevice;
import com.github.sarxos.webcam.ds.buildin.WebcamDefaultDriver;
import com.github.sarxos.webcam.ds.ipcam.IpCamDevice;
import com.github.sarxos.webcam.ds.ipcam.IpCamDeviceRegistry;
import com.github.sarxos.webcam.ds.ipcam.IpCamDriver;
import com.github.sarxos.webcam.ds.ipcam.IpCamMode;

import de.frankmuenster.jameica.qrscanner.QrCodeScanner;
import de.frankmuenster.jameica.qrscanner.WebcamWatchdog;
import de.frankmuenster.jameica.qrscanner.gui.parts.ImageLabelPart;
import de.frankmuenster.jameica.qrscanner.ipcam.DroidCamDevice;
import de.frankmuenster.jameica.qrscanner.ipcam.IphoneIpCameraCamDevice;
import de.willuhn.jameica.gui.input.TextAreaInput;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Level;
import de.willuhn.logging.Logger;
import de.willuhn.util.I18N;

/**
 * Die Klasse erstellt einen Hintergrund-task, der die Default-Webcam ffnet und solange scannt bis ein QR-Code gefunden
 * wurde oder der Task durch {@link QrCodeScannerThread#interrupt()} unterbrochen wird.
 *
 * @author Frank Mnster
 */
@SuppressWarnings("PMD.GodClass")
public class QrCodeScannerThread extends Thread {

    /**
     * i18n bersetzer aus unseren Settings
     */
    private final static I18N I18N = Application.getPluginLoader().getPlugin(QrCodeScanner.class).getResources()
            .getI18N();

    /**
     * Mindestbreite fr die Webcam-Auflsung
     */
    private static final int MIN_SIZE_WIDTH = 640;

    /**
     * Preferred dimension for webcam resolution
     */
    private static final Dimension MAX_DIMENSION = new Dimension(2560, 1440);

    /**
     * Die Listener fr einen QrCodeFound-Event
     */
    private transient final Set<QrCodeFoundListener> listeners = new CopyOnWriteArraySet<>();

    /**
     * Wurde der Thread von auen unterbrochen
     */
    private transient boolean interrupted;

    /**
     * TextArea fr Webcam Infos
     */
    private transient TextAreaInput webcamInfo;

    /**
     * Das Label fr das Webcambild
     */
    private transient ImageLabelPart imgLabel;

    /**
     * Der Name der Webcam, die benutzt werden soll
     */
    private transient String webcamName;

    /**
     * Der gefundene QR-Code als String
     */
    private transient String result;

    /**
     * Container fr den WebcamInfoText, da dieser in einem anderem Thread bentigt wird.
     */
    private transient String webcamInfoText;

    /**
     * Erstellt einen neuen QrCodeScanner Thread
     *
     * @param imgLabel
     *            das ImageLabelPart in, dem das Webcam-Bild dargestellt werden soll.
     * @param webcamInfo
     *            Textbereich fr WebCam Infos.
     * @param webcamName
     *            der Name der Webcam, die benutzt werden soll.
     * @throws NullArgumentException
     *             , wenn <b>imgLabel</b> <i>null</i> ist.
     */
    public QrCodeScannerThread(final ImageLabelPart imgLabel, final TextAreaInput webcamInfo, final String webcamName)
            throws NullArgumentException {

        super();
        if (imgLabel == null) {
            throw new NullArgumentException("imgLabel"); //$NON-NLS-1$
        } else if (webcamInfo == null) {
            throw new NullArgumentException("webcamInfo"); //$NON-NLS-1$
        } else if (webcamName == null) {
            throw new NullArgumentException("webcamName"); //$NON-NLS-1$
        } else {
            this.webcamInfo = webcamInfo;
            this.imgLabel = imgLabel;
            this.webcamName = webcamName;
            this.result = StringUtils.EMPTY;
        }
    }

    /**
     * Fgt eine Listener hinzu
     *
     * @param listener
     *            der hinzugefgt werden soll
     */
    public final void addListener(final QrCodeFoundListener listener) {

        this.listeners.add(listener);
    }

    /**
     * @see java.lang.Thread#interrupt()
     */
    @Override
    public void interrupt() {

        this.interrupted = true;
    }

    /**
     * @see java.lang.Thread#isInterrupted()
     */
    @Override
    public boolean isInterrupted() {

        return this.interrupted;
    }

    /**
     * Entfernt einen Listener aus der Liste
     *
     * @param listener
     *            der entfernt werden soll
     */
    public final void removeListener(final QrCodeFoundListener listener) {

        this.listeners.remove(listener);
    }

    /**
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {

        final Webcam webcam = this.getWebcamDevice(this.webcamName);
        if (webcam == null) {
            this.imgLabel.setText(QrCodeScannerThread.I18N.tr("Webcam (" + this.webcamName + ") nicht gefunden!"));
            if (Logger.isLogging(Level.INFO)) {
                Logger.info("Webcam (" + this.webcamName + ") nicht gefunden!");
            }
        } else {

            final Dimension dimension = this.getHighestResolution(webcam);

            if (dimension != null) {
                Logger.debug("Select Size: " + dimension.toString()); //$NON-NLS-1$
                webcam.setViewSize(dimension);
            }
            // start Loop
            if (!this.isInterrupted()) {
                this.loop(webcam);
            }
            this.closeCam(webcam);
        }
        Logger.debug("Lade Composite Treiber");
        Webcam.resetDriver();
        Webcam.setDriver(WebcamWatchdog.getDriver());
    }

    private void closeCam(final Webcam webcam) {

        if (webcam != null) {
            Logger.debug("Schliesse Webcam...");
            if (webcam.getLock().isLocked()) {
                Logger.debug(webcam.getLock().getLockFile().getAbsolutePath());
            }
            webcam.getLock().unlock();
            if (webcam.isOpen()) {
                webcam.close();
            }
        }
    }

    private WebcamDevice getDeviceByName(final String name) {

        WebcamDevice webcamDevice = null;
        final List<WebcamDevice> webcams = Webcam.getDriver().getDevices();
        for (final WebcamDevice webcamDev : webcams) {
            Logger.debug(webcamDev.toString());
            if (webcamDev.getName().equals(name)) {
                webcamDevice = webcamDev;
                break;
            }
        }
        return webcamDevice;
    }

    /**
     * Return the either HD resolution or the nearest highest one.
     *
     * @param webcam
     *            the webcam to check the resolution
     * @return the highest possible resolution (not higher than HD)
     */
    private Dimension getHighestResolution(final Webcam webcam) {

        // test HD
        webcam.setCustomViewSizes(QrCodeScannerThread.MAX_DIMENSION);
        return QrCodeScannerThread.MAX_DIMENSION;

        // final Dimension dimension = null;

        // webcam.setViewSize(QrCodeScannerThread.MAX_DIMENSION);
        // webcam.open();
        // final Dimension realResolution = webcam.getViewSize();
        //
        // if (realResolution.equals(QrCodeScannerThread.MAX_DIMENSION)) {
        // Logger.debug("HD is possible...");
        // dimension = realResolution;
        // } else {
        // // get max. resolution
        // final Dimension[] dimensions = webcam.getViewSizes();
        //
        // for (int i = 0; i < dimensions.length; i++) {
        // dimension = dimensions[i];
        // Logger.trace("Size: " + dimension.toString()); //$NON-NLS-1$
        // if (dimension.getWidth() >= QrCodeScannerThread.MIN_SIZE_WIDTH) {
        // break;
        // }
        // }
        // }
        // webcam.close();
        // return dimension;
    }

    private Webcam getWebcamDevice(final String name) {

        Webcam webcam = null;
        final WebcamDevice webcamDevice = this.getDeviceByName(name);
        if (webcamDevice != null) {
            Webcam.resetDriver();
            // USB Cam
            if (webcamDevice instanceof WebcamDefaultDevice) {
                Logger.debug("WebcamDefaultDevice");
                Webcam.setDriver(WebcamDefaultDriver.class);
                webcam = Webcam.getWebcamByName(name);
            } else if (webcamDevice instanceof IpCamDevice) {
                webcam = this.handleIpWebcam(webcamDevice);
            }
        }
        return webcam;
    }

    private void handleExceptions(final Exception exception) {

        if (exception.getClass() == TimeoutException.class) {
            this.imgLabel.setText(QrCodeScannerThread.I18N.tr("Es konnte keine Webcam gefunden werden..."));
            Logger.info("Keine untersttze Webcam gefunden");
        } else if (exception.getClass() == IllegalArgumentException.class) {
            this.imgLabel.setText(QrCodeScannerThread.I18N.tr("Auflsung wird nicht untersttzt!"));
            Logger.warn("Auflsung wird von Webcam nicht untersttzt");
        } else if (exception instanceof WebcamException
                && exception.getMessage().equals("Cannot get initial image from IP camera device DroidCam")) {
            this.imgLabel.setText(QrCodeScannerThread.I18N
                    .tr("Kann DroidCam nicht starten!\n\nGreift schon eine andere Applikation/Webseite darauf zu?"));
            Logger.warn("Kann DroidCam nicht starten! Evtl. greift schon eine andere Applikation/Webseite darauf zu.");
        } else {
            this.imgLabel.setText(QrCodeScannerThread.I18N.tr("Schwerwiegender Fehler"));
            Logger.error("Ein Fehler ist wrend der Webcamabfragen aufgetreten. Details im debug-log");
            Logger.debug("WebcamError: " + exception.toString());
        }
        try {
            Thread.sleep(3000);
        } catch (final InterruptedException e1) {
        }
    }

    private Webcam handleIpWebcam(final WebcamDevice webcamDevice) {

        Webcam webcam;
        // IP Cam
        Logger.debug("IpCamDevice Type: " + webcamDevice.getClass().getSimpleName());
        Webcam.setDriver(IpCamDriver.class);
        if (IpCamDeviceRegistry.isRegistered((IpCamDevice) webcamDevice)) {
            Logger.debug("Unregister " + IpCamDeviceRegistry.unregister(webcamDevice.getName()));
        }
        final URL url = ((IpCamDevice) webcamDevice).getURL();
        final IpCamMode mode = ((IpCamDevice) webcamDevice).getMode();
        final String name = ((IpCamDevice) webcamDevice).getName();
        webcamDevice.dispose();
        IpCamDevice cam;
        String notOnlineText;
        if (webcamDevice instanceof DroidCamDevice) {
            cam = new DroidCamDevice(url);
            notOnlineText = "\nDie DroidCam App unter " + url + " ist nicht online.\n"
                    + "Bitte URL berprfen, bzw DroidCam auf Android Handy starten.\n\n Warte...";
        } else if (webcamDevice instanceof IphoneIpCameraCamDevice) {
            cam = new IphoneIpCameraCamDevice(url);
            notOnlineText = "\nDie IP Camera App ist nicht online.\n"
                    + "Bitte APP auf dem I-Handy neu starten.\n\n Warte...";
        } else {
            cam = new IpCamDevice(name, url, mode);
            notOnlineText = "\nDie IP Kamera ist nicht online.\n"
                    + "Bitte APP auf dem I-Handy neu starten.\n\n Warte...";
        }
        IpCamDeviceRegistry.register(cam);
        webcam = Webcam.getWebcamByName(name);
        boolean isOnline = ((IpCamDevice) webcam.getDevice()).isOnline();
        Logger.debug("Online: " + isOnline);
        // Ist DoidCam gestartet?
        if (!isOnline) {
            this.imgLabel.setText(QrCodeScannerThread.I18N.tr(notOnlineText));
        }
        while (!isOnline && !this.isInterrupted()) {
            isOnline = ((IpCamDevice) webcam.getDevice()).isOnline();
            try {
                Thread.sleep(300);
            } catch (final InterruptedException e) {
            }
        }
        return webcam;
    }

    private void loop(final Webcam webcam) {

        String qrCode = null;
        try {
            // Informiere Benutzer, das Webcam geoeffnet wird
            this.imgLabel.setText(QrCodeScannerThread.I18N.tr("ffne Webcam..."));
            if (!webcam.isOpen()) {
                // webcam.dispose();
                webcam.open();
            }
            Logger.debug("Webcam Auflsung: " + webcam.getViewSize().toString());
            // Update Webcam Info
            this.webcamInfoText = "Test";
            if (webcam != null) {
                if (webcam.getViewSize().width < QrCodeScannerThread.MIN_SIZE_WIDTH) {
                    this.webcamInfoText = "Achtung: Die Webcam hat eine geringe (" + webcam.getViewSize().width + "x"
                            + webcam.getViewSize().height + ") Auflsung. Kleine QR-Codes werden "
                            + "eventuell nicht erkannt!";
                } else {
                    this.webcamInfoText = "Webcam Auflsung: " + webcam.getViewSize().width + "x"
                            + webcam.getViewSize().height;
                }
            }
            Display.getDefault().asyncExec(new Runnable() {

                /**
                 * @see java.lang.Runnable#run()
                 */
                @Override
                public void run() {

                    QrCodeScannerThread.this.webcamInfo.setValue(QrCodeScannerThread.this.webcamInfoText);
                }
            });
            final Point size = new Point(this.imgLabel.getSize().x, this.imgLabel.getSize().y);
            final ScanQrImage scanQrImage = new ScanQrImage(webcam, size);
            while (qrCode == null && !this.isInterrupted()) {
                // Setze Bild
                this.imgLabel.setImage(scanQrImage.getImage());
                // Checke QRCode
                qrCode = scanQrImage.getQrDecode();
                Thread.sleep(50);
            }
            this.closeCam(webcam);
            // set result String in case QR-Code was found
            if (qrCode != null) {
                Logger.debug("QR-Code decoded successfully!");
                this.result = qrCode;
                this.notifyListeners();
            }
        } catch (final Exception e) { // NOPMD
            this.handleExceptions(e);
        }
    }

    /**
     * Benachrichtigen der registrierten Listeners mit bergabe des gefundenen Strings
     */
    private void notifyListeners() {

        for (final QrCodeFoundListener listener : this.listeners) {
            listener.notifyOfQrCodeFound(this.result);
        }
    }
}
